#include "gaussLU.hh"
#include "alglin.hh"

// sstream contains the definmition of object that 
// are string used like file
#include <sstream>

// stdexcept contains the definition of the runtime_error object
// to use with try/catch blocks
#include <stdexcept>

#include <iostream>
#include <cmath>

#ifndef ASSERT
  #define ASSERT(COND,MSG) \
    if ( !(COND) ) { std::ostringstream ost ; ost << MSG << '\n' ; throw std::runtime_error(ost.str()) ; }
#endif

namespace gaussLU {

  using namespace std ;

  // access the element A(i,j) using fortran address
  #define A(I,J) A[(I)+(J)*lda]

  valueType
  factorize( indexType const n,      /* Number of rows and columns of matrix A */
             valueType       A[],    /* Pointer to coefficients of A           */
             indexType const lda,    /* Leading dimension of A, Fortran style  */
             indexType       perm[], /* Permutation vector                     */
             valueType       work[] ) {

    /*
    // decomposes a double precision matrix by gaussian elimination
    // and estimates the condition of the matrix.
    //
    // use solve to compute solutions to linear systems.
    // 
    // input..
    //
    //   lda = declared row dimension of the array containing A.
    //   n   = order of the matrix.
    //   A   = matrix to be triangularized.
    //
    // output..
    //
    //   A  contains an upper triangular matrix U and a permuted
    //      version of a lower triangular matrix  i-l  so that
    //      (permutation matrix)*a = l*u .
    //
    //   cond = an estimate of the condition of A.
    //          For the linear system A*x = b, changes in A and b
    //          may cause changes  cond  times as large in x.
    //          if cond+1.0 == cond, A is singular to working precision.
    //          cond is set to 1.0d+32 if exact singularity is detected.
    //
    //   perm      = the perm vector.
    //   perm[k]   = the index of the k-th perm row
    //   perm[n-1] = (-1)**(number of interchanges)
    //
    //   work space.. 
    //     the vector  work  must be declared and included in the call.
    //     Its input contents are ignored.
    //     Its output contents are usually unimportant.
    //
    //   the determinant of a can be obtained on output by
    //   det(A) = perm[n-1] * A(1,1) * A(2,2) * ... * A(n,n).
    */

    perm[n-1] = 1 ;
    if ( n == 1 ) { // 1-by-1
      valueType cond = 1 ;
      if ( A[0] == 0 ) cond = 1E32 ;
      return cond ;
    }

    // compute 1-norm of a
    valueType anorm = 0 ;
    for ( indexType j = 0 ; j < n ; ++j ) {
      valueType t = 0 ;
      for ( indexType i = 0 ; i < n ; ++i )
        t += abs(A(i,j)) ;
      if ( t > anorm ) anorm = t ;
    }

    // Gaussian elimination with partial pivoting
    indexType nm1 = n-1 ;
    for ( indexType k = 0 ; k < nm1 ; ++k ) {
      indexType kp1 = k+1 ;

      // find perm
      indexType m = k ;
      for ( indexType i = kp1 ; i < n ; ++i )
        if ( abs(A(i,k)) > abs(A(m,k)) )
          m = i ;

      perm[k] = m ;
      if ( m != k ) perm[n-1] *= -1 ; // update the sign of the permutation

      std::swap( A(m,k), A(k,k) ) ;

      // skip step if perm is zero
      valueType t = A(k,k) ;
      if ( t != 0 ) { // check if pivot is != 0 otherwise matrix is singular
        // compute multipliers
        for ( indexType i = kp1 ; i < n ; ++i )
          A(i,k) *= (-1/t) ;

        // interchange and eliminate by columns
        for ( indexType j = kp1 ; j < n ; ++j ) {
          std::swap( A(m,j), A(k,j) ) ;
          t = A(k,j) ;
          if ( t != 0 )
            for ( indexType i = kp1 ; i < n ; ++i )
              A(i,j) += A(i,k)*t ;
        }
      }
    }
    
    /*
    // cond = (1-norm of a)*(an estimate of 1-norm of a-inverse)
    // estimate obtained by one step of inverse iteration for the
    // small singular vector. This involves solving two systems
    // of equations, (A-transpose)*y = e  and A*z = y  where  e
    // is a vector of +1 or -1 chosen to cause growth in y.
    // estimate = (1-norm of z)/(1-norm of y)
    */

    // solve (A-transpose)*y = e

    // solve (U-transpose)z = e
    for ( indexType k = 0 ; k < n ; ++k ) {
      valueType t = 0 ;
      for ( indexType i = 0 ; i < k ; ++i ) t += A(i,k)*work[i] ;   
      if ( A(k,k) == 0 ) return 1E32 ;
      valueType ek = t < 0 ? -1 : +1 ;
      work[k] = -(ek + t)/A(k,k) ;
    }
    
    // solve (L-transpose)y = z
    indexType k = nm1 ;
    do { 
      --k ;
      valueType t = 0 ;
      for ( indexType i = k+1 ; i < n ; ++i )
        t += A(i,k)*work[i] ;   
      work[k] += t ;
      indexType m = perm[k] ;
      if ( m != k ) std::swap(work[k],work[m]) ;
    } while ( k > 0 ) ;

    valueType ynorm = 0 ;
    for ( indexType i = 0 ; i < n ; ++i )
      ynorm += abs(work[i]) ;

    // solve A*z = y
    solve(n, A, lda, perm, work ) ;

    valueType znorm = 0 ;
    for ( indexType i = 0 ; i < n ; ++i )
      znorm += abs(work[i]) ;

    // estimate condition
    valueType cond = anorm*znorm/ynorm ;
    if ( cond < 1 ) cond = 1 ;
    return cond ;


    // exact singularity
  exact_singularity:
    cond = 1E32 ;
  }


  valueType
  factorBlock( indexType const n,      /* Number of rows and columns of matrix A */
               valueType       A[],    /* Pointer to coefficients of A           */
               indexType const lda,    /* Leading dimension of A, Fortran style  */
               indexType       perm[], /* Permutation vector                     */
               valueType       work[] ) {

    /*
    // decomposes a double precision matrix by gaussian elimination
    // and estimates the condition of the matrix.
    //
    // use solve to compute solutions to linear systems.
    // 
    // input..
    //
    //   lda = declared row dimension of the array containing A.
    //   n   = order of the matrix.
    //   A   = matrix to be triangularized.
    //
    // output..
    //
    //   A  contains an upper triangular matrix U and a permuted
    //      version of a lower triangular matrix  i-l  so that
    //      (permutation matrix)*a = l*u .
    //
    //   cond = an estimate of the condition of A.
    //          For the linear system A*x = b, changes in A and b
    //          may cause changes  cond  times as large in x.
    //          if cond+1.0 == cond, A is singular to working precision.
    //          cond is set to 1.0d+32 if exact singularity is detected.
    //
    //   perm      = the perm vector.
    //   perm[k]   = the index of the k-th perm row
    //   perm[n-1] = (-1)**(number of interchanges)
    //
    //   work space.. 
    //     the vector  work  must be declared and included in the call.
    //     Its input contents are ignored.
    //     Its output contents are usually unimportant.
    //
    //   the determinant of a can be obtained on output by
    //   det(A) = perm[n-1] * A(1,1) * A(2,2) * ... * A(n,n).
    */

    perm[n-1] = 1 ;
    if ( n == 1 ) { // 1-by-1
      valueType cond = 1 ;
      if ( A[0] == 0 ) cond = 1E32 ;
      return cond ;
    }

    // compute 1-norm of a
    valueType anorm = alglin::norm1( n, n, A, lda ) ;

    // Gaussian elimination with partial pivoting
    indexType nm1 = n-1 ;
    for ( indexType k = 0 ; k < nm1 ; ++k ) {
      indexType kp1 = k+1 ;

      // find perm
      indexType m = k + alglin::iamax( n-k, &A(k,k), 1 ) ;

      perm[k] = m ;
      if ( m != k ) { // scambia righe
        perm[n-1] *= -1 ;
        alglin::swap(n-k, &A(k,k), lda, &A(m,k), lda ) ;
      }

      // skip step if perm is zero
      valueType t = A(k,k) ;
      if ( t != 0 ) {
        // compute multipliers
        alglin::scal( n-kp1, (-1/t), &A(kp1,k), 1 ) ;
        alglin::ger( n-kp1, n-kp1, 1,
                     &A(kp1,k),    1,
                     &A(k,kp1),    lda,
                     &A(kp1,kp1),  lda ); 
      }
    }
    
    /*
    // cond = (1-norm of a)*(an estimate of 1-norm of a-inverse)
    // estimate obtained by one step of inverse iteration for the
    // small singular vector. This involves solving two systems
    // of equations, (A-transpose)*y = e  and A*z = y  where  e
    // is a vector of +1 or -1 chosen to cause growth in y.
    // estimate = (1-norm of z)/(1-norm of y)
    */

    // solve (A-transpose)*y = e

    // solve (U-transpose)z = e
    for ( indexType k = 0 ; k < n ; ++k ) {
      valueType t = 0 ;
      for ( indexType i = 0 ; i < k ; ++i ) t += A(i,k)*work[i] ;   
      if ( A(k,k) == 0 ) return 1E32 ;
      valueType ek = t < 0 ? -1 : +1 ;
      work[k] = -(ek + t)/A(k,k) ;
    }
    
    // solve (L-transpose)y = z
    indexType k = nm1 ;
    do { 
      --k ;
      valueType t = 0 ;
      for ( indexType i = k+1 ; i < n ; ++i )
        t += A(i,k)*work[i] ;   
      work[k] += t ;
      indexType m = perm[k] ;
      if ( m != k ) std::swap(work[k],work[m]) ;
    } while ( k > 0 ) ;

    valueType ynorm = 0 ;
    for ( indexType i = 0 ; i < n ; ++i )
      ynorm += abs(work[i]) ;

    // solve A*z = y
    solve(n, A, lda, perm, work ) ;

    valueType znorm = 0 ;
    for ( indexType i = 0 ; i < n ; ++i )
      znorm += abs(work[i]) ;

    // estimate condition

    valueType cond = anorm*znorm/ynorm ;
    if ( cond < 1 ) cond = 1 ;
    return cond ;


    // exact singularity
  exact_singularity:
    cond = 1E32 ;
  }

  #define L(I,J) LU[(I)+(J)*lda]
  #define U(I,J) LU[(I)+(J)*lda]

  void
  solve( indexType const n,      /* Number of rows and columns of matrix A */
         valueType const LU[],   /* Pointer to coefficients of LU          */
         indexType const lda,    /* Leading dimension of LU, Fortran style */
         indexType const perm[], /* Permutation vector                     */
         valueType       b[]     /* Pointer to coefficients of vector b    */  ) {

    /*
    // Solution of linear system, A*x = b.
    // do not use if decomp has detected singularity.
    // 
    // input..
    //   lda  = declared row dimension of array containing a .
    //   n    = order of matrix.
    //   LU   = triangularized matrix obtained from decomp .
    //   b    = right hand side vector.
    //   perm = pivot vector obtained from decomp .
    //
    // output..
    //   b = solution vector, x.
    */
    
    if ( n == 1 ) {
      b[0] /= LU[0] ;
      return ;
    }

    // forward elimination Lz = b
    indexType nm1 = n-1 ;
    for ( indexType k = 0 ; k < nm1 ; ++k ) {
      indexType m = perm[k] ;
      if ( m != k ) std::swap( b[m], b[k] ) ;
      valueType t = b[k] ;
      for ( indexType i = k+1 ; i < n ; ++i )
        b[i] += L(i,k)*t ;
    }
    // back substitution Ux = z
    indexType k = n ;
    do {
      --k ;
      b[k] /= U(k,k) ;
      valueType t = b[k] ;
      for ( indexType i=0 ; i < k ; ++i )
        b[i] -= U(i,k)*t ;
    } while ( k > 0 ) ;
  }

  void
  solveTranspose( indexType const n,      /* Number of rows and columns of matrix A */
                  valueType const LU[],   /* Pointer to coefficients of LU          */
                  indexType const lda,    /* Leading dimension of LU, Fortran style */
                  indexType const perm[], /* Permutation vector                     */
                  valueType       b[]     /* Pointer to coefficients of vector b    */  ) {

    /*
    // Solution of linear system, (A-transpose)*x = b.
    // do not use if decomp has detected singularity.
    // 
    // input..
    //   lda  = declared row dimension of array containing a .
    //   n    = order of matrix.
    //   LU   = triangularized matrix obtained from decomp .
    //   b    = right hand side vector.
    //   perm = pivot vector obtained from decomp .
    //
    // output..
    //   b = solution vector, x.
    */
    
    if ( n == 1 ) {
      b[0] /= LU[0] ;
      return ;
    }

    // solve (U-transpose)z = e
    for ( indexType k = 0 ; k < n ; ++k ) {
      valueType t = 0 ;
      for ( indexType i = 0 ; i < k ; ++i ) t += U(i,k)*b[i] ;   
      b[k] = (b[k]-t)/U(k,k) ;
    }

    // solve (L-transpose)y = z
    indexType k = n-1 ;
    do { 
      --k ;
      valueType t = 0 ;
      for ( indexType i = k+1 ; i < n ; ++i )
        t += L(i,k)*b[i] ;   
      b[k] += t ;
      indexType m = perm[k] ;
      if ( m != k ) std::swap(b[k],b[m]) ;
    } while ( k > 0 ) ;
  }
}
