#include "recursiveLU.hh"
#include "alglin.hh"

#include <sstream>
#include <iostream>
#include <cmath>

#ifndef ASSERT
  #define ASSERT(COND,MSG) \
    if ( !(COND) ) { std::ostringstream ost ; ost << MSG << '\n' ; throw std::runtime_error(ost.str()) ; }
#endif

using namespace std ;

namespace recursiveLU {

  #define A(I,J) A[(I)+(J)*lda]
  #define IPIV   (alglin::integer*)ipiv

  /*
  //
  //  +-------------+-------------+-------------+-------------+ 0
  //  |             |             |             |             |
  //  |             |             |             |             |
  //  |             |             |             |             |
  //  |             |             |             |             |
  //  |             |             |             |             |
  //  +-------------+=============+=============+-------------+ N0
  //  *            || \           |             ||            *
  //  *            ||   \    U1   |             ||            *
  //  *     C1     ||     \       |      A1     ||            *
  //  *            ||       \     |             ||     B1     *
  //  *            ||   L1    \   |             ||            *
  //  *            ||           \ |             ||            *
  //  +-------------+-------------+-------------+-------------+ N1
  //  *            ||             | \           ||            *
  //  *            ||             |   \    U2   ||            *
  //  *     C2     ||     A2      |     \       ||     B2     *
  //  *            ||             |       \     ||            *
  //  *            ||             |  L2     \   ||            *
  //  *            ||             |           \ ||            *
  //  +-------------+=============+=============+-------------+ N2
  //  |             |             |             |             |
  //  |             |             |             |             |
  //  |             |             |             |             |
  //  |             |             |             |             |
  //  |             |             |             |             |
  //  |             |             |             |             |
  //  +-------------+-------------+-------------+-------------+ numRows
  //  0            N0             N1            N2         numCols
  //
  */

  indexType
  factorize( indexType const N0,
             indexType const N2,
             indexType const numRows,
             indexType const numCols,
             valueType       A[],       /* Pointer to coefficients of A           */
             indexType const lda,       /* Leading dimension of A, Fortran style  */
             indexType       ipiv[] ) { /* Permutation vector                     */

    indexType N1 = (N2+N0)/2  ;

    valueType * LU1 = A   + N0*(1+lda) ;
    valueType * LU2 = A   + N1*(1+lda) ;
    valueType * A1  = LU1 + (N1-N0)*lda ;
    valueType * A2  = LU1 + N1-N0 ;

    if ( N2-N0 < 80 ) {    
      indexType res = alglin::getrf( N2-N0, N2-N0, LU1, lda, IPIV + N0 ) ; // do factorization
      for ( indexType i = N0 ; i < N2 ; ++i ) ipiv[i] += N0 ;
      alglin::swaps( numCols-N2, A + N2*lda, lda, N0, N2-1, IPIV, 1 ) ; // apply row permutation to B1
      alglin::swaps( N0,         A,          lda, N0, N2-1, IPIV, 1 ) ; // apply row permutation to C1
      return res ;
    }

    factorize( N0, N1, numRows, numCols, A, lda, ipiv ) ; // do factorization block LU1

    //
    // +---------+---------+
    // | \    U1 |         |
    // |    \    | L(-1)A1 |
    // |  L1   \ |         |
    // |=========|=========|
    // |         |         |
    // |A2 U^(-1)|    CC   |
    // |         |         |
    // +---------+---------+
    //
    alglin::trsm( "Left",  "Lower", "No Transpose", "Unit",     N1-N0, N2-N1, 1.0, LU1, lda, A1, lda ) ; // A1 <- L^(-1)A1
    alglin::trsm( "Right", "Upper", "No Transpose", "Non Unit", N2-N1, N1-N0, 1.0, LU1, lda, A2, lda ) ; // A1 <- L^(-1)A1
    //
    // +---------+---------+
    // |\    U   |         |
    // |   \     |L^(-1)A1 |
    // |   L   \ |         |
    // |=========|=========|
    // |         |         |
    // |A2 U^(-1)|    CC*  | CC* = CC - A2 U^(-1)(L^(-1)A1)
    // |         |         |
    // +---------+---------+
    //
    alglin::gemm( "No Transpose", "No Transpose", N2-N1, N2-N1, N1-N0,
                  -1.0, A2,  lda, A1, lda,
                   1.0, LU2, lda ) ; // CC - A2 L^(-1)A1

    factorize( N1, N2, numRows, numCols, A, lda, ipiv ) ; // do factorization block LU2

    return 0 ;
  }

  indexType
  factorize( indexType const m,         /* Number of rows of matrix A             */
             indexType const n,         /* Number of columns of matrix A          */
             valueType       A[],       /* Pointer to coefficients of A           */
             indexType const lda,       /* Leading dimension of A, Fortran style  */
             indexType       ipiv[] ) { /* Permutation vector                     */

    return factorize( 0, min(m,n), m, n, A, lda, ipiv ) ;
  }

}
